
package macrolanguage;

import java.awt.*;
import java.awt.event.*;
import java.util.Hashtable;
import macrolanguage.ProjectSettings;

/**
 * the ComponentWrapper class wrapps a component with a panel
 * allowing it to be manipulated in the FreeFormEdit with greater
 * freedom.
 * The need for the ComponentWrapper arose when TextArea and similar
 * classes would not resize since they block some of the events which
 * should be sent by them.
**/
public class ComponentWrapper extends Panel implements MouseListener, MouseMotionListener
{
		private Point lastMouseLocation = null;

		// the margin bettween the components edge and the container.
		private static int MARGIN=4;

		// the component contained inside the panel.
		private Component internalComponent;


		/**
		 * Initialize the component wrapper to wrap internalComponent.
		**/
		public ComponentWrapper(Component internalComponent)
		{
				super();
				
				// the null layout means no layout manager which means
				// components will not rearange them selves
				setLayout(null);
				
				this.internalComponent = internalComponent;

				// prevent the Panel from resizeing the internalComponent
				// when it is layed out.
				super.add(internalComponent);

				// handle all of the events brodcasted by the intrenal component
				// and this component.
				addMouseListener(this);
				internalComponent.addMouseListener(this);
				addMouseMotionListener(this);
				internalComponent.addMouseMotionListener(this);
		}
		/**
		 * called whenever a component gets dragged and performas the drag
		 * on the internal Component as well.
		**/
		private void componentDragged(java.awt.event.MouseEvent ev)
		{
				// remove the selection from the selected component to avoid
				// a trail.
				ProjectSettings.currentFreeFormEdit.setSelection(ProjectSettings.currentFreeFormEdit);

				// get the position of the event.
				Point pos = ev.getPoint();

				// stop listening to mouse events for a short while.
				removeMouseMotionListener(this);

				ProjectSettings.dragMode = true;

				if(lastMouseLocation == null)
				{
						lastMouseLocation = pos;
				}
				else
				{
						setLocation(    pos.x - lastMouseLocation.x + getLocation().x,
										pos.y - lastMouseLocation.y + getLocation().y);
						lastMouseLocation = null;
				}

				// resume listening to mouse events.
				addMouseMotionListener(this);
		}
		/**
		 * this method is called whenever a component gets resized
		 * and it handles the resizeing for the internal component too.
		**/
		private void componentResized(java.awt.event.MouseEvent ev)
		{
				// get the position of the event.
				Point pos = ev.getPoint();

				Component c = (Component)ev.getSource();

				c.removeMouseMotionListener(this);

				if ((pos.x < ProjectSettings.MINIMUM_COMPONENT_WIDTH) ||
					(pos.y < ProjectSettings.MINIMUM_COMPONENT_HEIGHT))
						return;

				// if we have a horizontal resize mouse then we should
				// resize the y azis, if not we should leave it as before.
				if(((ProjectSettings.resizeMode & ProjectSettings.RESIZE_UP) == 0) && ((ProjectSettings.resizeMode & ProjectSettings.RESIZE_DOWN) == 0))
					if(c instanceof ComponentWrapper)
						pos.y = ((ComponentWrapper)c).getComponent().getSize().height;
					else
						pos.y = c.getSize().height;

				// if we have a vertical resize mouse then we should
				// resize the x azis, if not we should leave it as before.
				if(((ProjectSettings.resizeMode & ProjectSettings.RESIZE_LEFT) == 0) && ((ProjectSettings.resizeMode & ProjectSettings.RESIZE_RIGHT) == 0))
					if(c instanceof ComponentWrapper)
							pos.x = ((ComponentWrapper)c).getComponent().getSize().width;
						else
							pos.x = c.getSize().width;

				// check if the x asis is on the grid.
				if(pos.x % ProjectSettings.gridSize != 0)
						pos.x = (pos.x / ProjectSettings.gridSize) * (pos.x % ProjectSettings.gridSize);

				// check if the y asis is on the grid.
				if(pos.y % ProjectSettings.gridSize != 0)
						pos.y = (pos.y / ProjectSettings.gridSize) * (pos.y % ProjectSettings.gridSize);

				setSize(pos.x, pos.y);

				c.addMouseMotionListener(this);
		}
		/**
		 * remove the selected marking on this component.
		**/
		public void deselect()
		{
				setBackground(SystemColor.window);
		}
		/**
		 * get the internal component.
		**/
		public Component getComponent()
		{
				return(internalComponent);
		}
		/**
		 * required by the MouseListener interface.
		 * When a mouse is clicked that could mean 2 things:
		 * 1. Selection of a component on the surface.
		 * 2. Dropping a new component onto the surface.
		**/
		public void mouseClicked(java.awt.event.MouseEvent ev)
		{
				// if we are in the midst of placeing a new component onto
				// the free edit surface.
				if(ProjectSettings.componentToPlace != null)
				{
						try
						{
								// set the mouse cursor to the original cursor.
								ProjectSettings.currentFreeFormEdit.setCursor(Cursor.getDefaultCursor());

								ComponentWrapper wrapper = new ComponentWrapper(ProjectSettings.componentToPlace);

								// add the new component to the correct location.
								ProjectSettings.currentFreeFormEdit.add(wrapper, ev.getPoint().x, ev.getPoint().y);

								// set the size of the component to the default size for all the components.
								wrapper.setSize(ProjectSettings.DEFAULT_COMPONENT_HEIGHT, ProjectSettings.DEFAULT_COMPONENT_WIDTH);

								// set the new component to be the selected component.
								ProjectSettings.currentFreeFormEdit.setSelection(wrapper);

								// mark the fact that the component was placed.
								ProjectSettings.componentToPlace = null;
								return;
						}
						catch(NotAContainerException e)
						{
								System.out.println("Not a container. A component can only be placed in a container.");
						}
				}

				// set the selected component to be the component which
				// generated the mouse click event.
				ProjectSettings.currentFreeFormEdit.setSelection((Component)ev.getSource());
		}
		/**
		 * This method is a part of the MouseMotionListener interface and is
		 * called whenever a mouse is dragged.
		 * This method abstracts the dragging and dropping of components
		 * on the free form surface.
		**/
		public void mouseDragged(java.awt.event.MouseEvent ev)
		{
				// check if the dragging is the concequence of a resize
				// operation or move operation.
				if(ProjectSettings.resizeMode == ProjectSettings.NO_RESIZE)
						componentDragged(ev);
				else
						componentResized(ev);
		}
		/**
		 * required by the MouseMotionListener interface.
		**/
		public void mouseEntered(java.awt.event.MouseEvent ev)
		{
		}
		/**
		 * required by the MouseMotionListener interface.
		**/
		public void mouseExited(java.awt.event.MouseEvent ev)
		{
		}
		/**
		 * required by the MouseMotionListener interface.
		 * used to determin when to diplay the resize mouse pointer on the components.
		 * The resize mode is used not only for the look of the mouse pointer
		 * it is used to determin whether to resize a component or move it
		 * when dragging.
		**/
		public void mouseMoved(java.awt.event.MouseEvent ev)
		{
				// we can't resize components while placeing new objects.
				if (ProjectSettings.componentToPlace != null) return;

				// removing indication of the previos resize mode.
				ProjectSettings.resizeMode = ProjectSettings.NO_RESIZE;

				// the component which generated the event.
				Component comp = (Component)ev.getSource();

				// the location of the event.
				Point pos = ev.getPoint();

				int num;

				// if this component generated the event or if the internal
				// component generated the event.
				if(comp instanceof ComponentWrapper)
				{
						// If the x is bigger than the width of the internal 
						// component then allow a right resizeing.
						if(pos.x >= getComponent().getSize().width)
								ProjectSettings.resizeMode = ProjectSettings.RESIZE_RIGHT;

						// If the y is bigger than the height of the internal 
						// component then allow a down resizeing.
						if(pos.y >= getComponent().getSize().height)
								ProjectSettings.resizeMode = ProjectSettings.resizeMode | ProjectSettings.RESIZE_DOWN;

						num = ((Integer)ProjectSettings.resizePossibleStates.get(new Integer(ProjectSettings.resizeMode))).intValue();
				}
				else
				{
						//  if we are in the internal component then take into
						// consideration that a certin distance from the rim of
						// the component must exist for the resize to be valid.
						// That size is ProjectSettings.RESIZE_ACTIVATION_BOUNDRY.
						if(pos.x > (comp.getSize().width - ProjectSettings.RESIZE_ACTIVATION_BOUNDRY))
								ProjectSettings.resizeMode = ProjectSettings.RESIZE_RIGHT;

						if(pos.y > (comp.getSize().height - ProjectSettings.RESIZE_ACTIVATION_BOUNDRY))
								ProjectSettings.resizeMode = ProjectSettings.resizeMode | ProjectSettings.RESIZE_DOWN;

						if(pos.x < ProjectSettings.RESIZE_ACTIVATION_BOUNDRY)
								ProjectSettings.resizeMode = ProjectSettings.resizeMode | ProjectSettings.RESIZE_LEFT;

						if(pos.y < ProjectSettings.RESIZE_ACTIVATION_BOUNDRY)
								ProjectSettings.resizeMode = ProjectSettings.resizeMode | ProjectSettings.RESIZE_UP;

						num = ((Integer)ProjectSettings.resizePossibleStates.get(new Integer(ProjectSettings.resizeMode))).intValue();
				}

				// get the number of the cursor to use.
				num = ((Integer)ProjectSettings.resizePossibleStates.get(new Integer(ProjectSettings.resizeMode))).intValue();

				// set the correct cursor. for the internal and external 
				// componets.
				ProjectSettings.currentFreeFormEdit.setCursor(new Cursor(num));
				setCursor(new Cursor(num));
				getComponent().setCursor(new Cursor(num));

				// prevent this event from being distributed to other listeners.
				ev.consume();
		}
		/**
		 * required by the MouseListener interface.
		**/
		public void mousePressed(java.awt.event.MouseEvent ev)
		{
		}
		/**
		 * required by the MouseListener interface.
		 * The drop operation when dragging is implemented here.
		**/
		public void mouseReleased(java.awt.event.MouseEvent ev)
		{
				// this method is in essence a close circle for the 
				// drag operation implemented in mouseDragged.
				if(ProjectSettings.dragMode && (lastMouseLocation != null))
				{
						setLocation(    ev.getPoint().x - lastMouseLocation.x + getLocation().x,
										ev.getPoint().y - lastMouseLocation.y + getLocation().y);
						lastMouseLocation = null;
						ProjectSettings.dragMode = false;
				}
				ProjectSettings.currentFreeFormEdit.repaint();
		}
		/**
		 * mark this component as selected.
		**/
		public void select()
		{
				setBackground(Color.black);
		}
		/**
		 * set the size of the component (and the container).
		 * accepts the new height and width of the component.
		**/
		public void setSize(int width, int height)
		{
				super.setSize(width + (MARGIN * 2), height + (MARGIN * 2));
				internalComponent.setSize(width, height);
				internalComponent.setLocation(MARGIN, MARGIN);
		}
		/**
		 * set the size of the component (and the container).
		**/
		public void setSize(Dimension size)
		{
				setSize(size.width, size.height);
		}
}